<?php

/**
 * This file is part of the MADIS - RGPD Management application.
 *
 * @copyright Copyright (c) 2018-2019 Soluris - Solutions Numériques Territoriales Innovantes
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 */

declare(strict_types=1);

namespace App\Tests\Application\Symfony\EventSubscriber\Doctrine;

use App\Application\Symfony\EventSubscriber\Doctrine\LinkCreatorSubscriber;
use App\Application\Symfony\Security\UserProvider;
use App\Application\Traits\Model\CreatorTrait;
use App\Domain\User\Model;
use App\Tests\Utils\ReflectionTrait;
use Doctrine\Common\EventSubscriber;
use Doctrine\ORM\EntityManagerInterface;
use Doctrine\ORM\Event\PrePersistEventArgs;
use PHPUnit\Framework\TestCase;
use Prophecy\PhpUnit\ProphecyTrait;
use Symfony\Component\Security\Core\Authentication\Token\SwitchUserToken;
use Symfony\Component\Security\Core\Authentication\Token\TokenInterface;
use Symfony\Component\Security\Core\Role\SwitchUserRole;
use Symfony\Component\Security\Core\User\UserInterface;

class LinkCreatorSubscriberTest extends TestCase
{
    use ReflectionTrait;
    use ProphecyTrait;

    /**
     * @var UserProvider
     */
    private $userProviderProphecy;

    /**
     * @var EntityManagerInterface
     */
    private $entityManagerProphecy;

    /**
     * @var LinkCreatorSubscriber
     */
    private $sut;

    protected function setUp(): void
    {
        $this->userProviderProphecy  = $this->prophesize(UserProvider::class);
        $this->entityManagerProphecy = $this->prophesize(EntityManagerInterface::class);

        parent::setUp();
    }

    protected function getSut(bool $linkAdmin = false): LinkCreatorSubscriber
    {
        return new LinkCreatorSubscriber(
            $this->userProviderProphecy->reveal(),
            $linkAdmin
        );
    }

    /**
     * Test instance of Subscriber.
     */
    public function testInstanceOf(): void
    {
        $this->assertInstanceOf(EventSubscriber::class, $this->getSut());
    }

    /**
     * Test getSubscribedEvents of current subscriber.
     */
    public function testGetSubscribedEvents(): void
    {
        $this->assertEquals(
            [
                'prePersist',
            ],
            $this->getSut()->getSubscribedEvents()
        );
    }

    /**
     * Test prePersist
     * Object has no CreatorTrait trait.
     *
     * @throws \Exception
     */
    public function testPrePersistNoCreatorTrait(): void
    {
        $object = new class {
        };
        $objectUses = \class_uses($object);
        $user       = new Model\User();
        $linkAdmin  = false;

        $tokenProphecy = $this->prophesize(TokenInterface::class);

        $tokenProphecy->getUser()->shouldBeCalled()->willReturn($user);
        $this->userProviderProphecy->getToken()->shouldBeCalled()->willReturn($tokenProphecy->reveal());

        $eventArgs = new PrePersistEventArgs($object, $this->entityManagerProphecy->reveal());

        $this->assertFalse(\in_array(CreatorTrait::class, $objectUses));
        $this->getSut($linkAdmin)->prePersist($eventArgs);
        // This test must only not return error
    }

    /**
     * Test prePersist
     * Object has CreatorTrait but user is not a User class.
     *
     * @throws \Exception
     */
    public function testPrePersistNoAuthenticatedUser(): void
    {
        $object     = new DummyLinkCreatorSubscriberTest();
        $objectUses = \class_uses($object);
        $user       = $this->prophesize(UserInterface::class)->reveal();
        $linkAdmin  = false;

        $tokenProphecy = $this->prophesize(TokenInterface::class);

        $tokenProphecy->getUser()->shouldBeCalled()->willReturn($user);
        $this->userProviderProphecy->getToken()->shouldBeCalled()->willReturn($tokenProphecy->reveal());
        $eventArgs = new PrePersistEventArgs($object, $this->entityManagerProphecy->reveal());

        $this->assertTrue(\in_array(CreatorTrait::class, $objectUses));
        $this->assertFalse($user instanceof Model\User);
        $this->getSut($linkAdmin)->prePersist($eventArgs);
        $this->assertNull($object->getCreator());
    }

    /**
     * Test prePersist
     * Object already has creator, then the connected user is not auto linked.
     */
    public function testPrePersistUserAlreadySet(): void
    {
        $object            = new DummyLinkCreatorSubscriberTest();
        $objectUses        = \class_uses($object);
        $alreadyLinkedUser = new Model\User();
        $object->setCreator($alreadyLinkedUser);
        $loggerUser = new Model\User();
        $linkAdmin  = false;

        $tokenProphecy = $this->prophesize(TokenInterface::class);

        $tokenProphecy->getUser()->shouldBeCalled()->willReturn($loggerUser);
        $this->userProviderProphecy->getToken()->shouldBeCalled()->willReturn($tokenProphecy->reveal());
        $eventArgs = new PrePersistEventArgs($object, $this->entityManagerProphecy->reveal());

        $this->assertTrue(\in_array(CreatorTrait::class, $objectUses));
        $this->assertEquals($alreadyLinkedUser, $object->getCreator());
        $this->getSut($linkAdmin)->prePersist($eventArgs);
        $this->assertEquals($alreadyLinkedUser, $object->getCreator());
    }

    /**
     * Test prePersist
     * Don't link admin as creator.
     *
     * @throws \Exception
     */
    public function testPrePersistNoLinkAdmin(): void
    {
        $object     = new DummyLinkCreatorSubscriberTest();
        $objectUses = \class_uses($object);
        $user       = new Model\User();
        $linkAdmin  = false;

        $tokenProphecy = $this->prophesize(TokenInterface::class);

        $tokenProphecy->getUser()->shouldBeCalled()->willReturn($user);
        $this->userProviderProphecy->getToken()->shouldBeCalled()->willReturn($tokenProphecy->reveal());
        $eventArgs = new PrePersistEventArgs($object, $this->entityManagerProphecy->reveal());

        $this->assertTrue(\in_array(CreatorTrait::class, $objectUses));
        $this->assertTrue($user instanceof Model\User);
        $this->getSut($linkAdmin)->prePersist($eventArgs);
        $this->assertEquals($user, $object->getCreator());
    }

    /**
     * Test prePersist
     * Link admin as creator.
     *
     * @throws \Exception
     */
    public function testPrePersistLinkAdmin(): void
    {
        $object     = new DummyLinkCreatorSubscriberTest();
        $objectUses = \class_uses($object);
        $user       = new Model\User();
        $admin      = new Model\User();
        $linkAdmin  = true;

        $tokenProphecy         = $this->prophesize(SwitchUserToken::class);
        $originalTokenProphecy = $this->prophesize(TokenInterface::class);

        $tokenProphecy->getUser()->shouldBeCalled()->willReturn($user);
        $tokenProphecy->getOriginalToken()->shouldBeCalled()->willReturn($originalTokenProphecy);
        $originalTokenProphecy->getUser()->shouldBeCalled()->willReturn($admin);

        $this->userProviderProphecy->getToken()->shouldBeCalled()->willReturn($tokenProphecy->reveal());
        $eventArgs = new PrePersistEventArgs($object, $this->entityManagerProphecy->reveal());

        $this->entityManagerProphecy->find(Model\User::class, $admin->getId()->toString())->shouldBeCalled()->willReturn($admin);

        $this->assertTrue(\in_array(CreatorTrait::class, $objectUses));
        $this->assertTrue($user instanceof Model\User);

        $this->getSut($linkAdmin)->prePersist($eventArgs);
        $this->assertEquals($admin, $object->getCreator());
    }

    /**
     * Test prePersist
     * Link admin as creator but there is no SwitchUserRole, then connected user is linked by default.
     *
     * @throws \Exception
     */
    public function testPrePersistLinkAdminNoSwitchUserRole(): void
    {
        $object     = new DummyLinkCreatorSubscriberTest();
        $objectUses = \class_uses($object);
        $user       = new Model\User();
        $linkAdmin  = true;

        $tokenProphecy = $this->prophesize(TokenInterface::class);

        $tokenProphecy->getUser()->shouldBeCalled()->willReturn($user);

        $this->userProviderProphecy->getToken()->shouldBeCalled()->willReturn($tokenProphecy->reveal());
        $eventArgs = new PrePersistEventArgs($object, $this->entityManagerProphecy->reveal());

        $this->assertTrue(\in_array(CreatorTrait::class, $objectUses));
        $this->assertTrue($user instanceof Model\User);

        $this->getSut($linkAdmin)->prePersist($eventArgs);
        $this->assertEquals($user, $object->getCreator());
    }
}

class DummyLinkCreatorSubscriberTest
{
    use CreatorTrait;
}
